# -*- coding: utf-8 -*-
"""
Created on Fri Mar 23 11:33:24 2018

@author: stuart
"""

#group tools for nanowire-plugin



############################
### Group specific tools ###
############################

from minio import Minio
import subprocess
import os
import tarfile
import traceback
import logging
import json
from os import environ
import inspect

import requests

import time


#import time
import sys
import datetime
import shutil
#import hashlib

from nanowire_plugin import send

#import the relavant version of urllib depending on the version of python we are
import urllib

#set up the logger globally
logger = logging.getLogger("nanowire-plugin")

logging.getLogger("urllib").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


#This is the worker that does the uploading
class GroupWorker(object):
    def __init__(self, function, minio_client, monitor_url, debug_mode):
        
        self.function = function
        self.minio_client = minio_client
        self.monitor_url = monitor_url
        self.debug_mode = debug_mode
        self.connect_max_retries=5
        self.new = True
        self.backoff = 0

            
    def run(self):
        while True:
            
            try:
                if 'REQUEST_TIMEOUT' in os.environ.keys():
                    time_out = int(os.environ['REQUEST_TIMEOUT'])
                else:
                    time_out = 5
    
                message = requests.get(os.environ['CONTROLLER_BASE_URI'] + '/v1/tasks/?pluginId=' + os.environ['PLUGIN_ID'] + '&pluginInstance=' + os.environ['POD_NAME'], timeout = time_out)
                
                code = message.status_code
                
            except:
                code = 500
            
            
            if code == 200:
                
                payload = json.loads(message.text)
                
                meta = payload['metadata']
                
                    
                try:
                    tar_url = meta['task']['metadata']['cacheURL']
                    found_tarball=True
                except:
                    tar_url = None
                    
    
                if meta['job']['workflow']['type'] != 'GROUP':
                    logger.warning("GROUP PLUGIN WAS SENT A NONE GROUP JOB")
                    
                else:
                    if tar_url != None:
                        try:
                            pull_and_extract_tarball(tar_url, '/cache')
                        except Exception as exp:
                            
                            if "COULD NOT FIND TARBALL AT" in str(exp):
                                found_tarball = False                    
            
            
                        if found_tarball:
                            read_obj = reader()
                            write_obj = writer(meta)
                            #************** There needs to be some way of getting the url before we hit this
                            
                            try:
                                #result = self.function(meta, jsonld, url)
                                result = run_group_function(self.function, read_obj, write_obj, meta)
                                
                            except Exception as exp:
                                if self.debug_mode > 0:
                                    result = str(traceback.format_exc())
                                    logger.info("THERE WAS A PROBLEM RUNNING THE MAIN FUNCTION: %s"%str(result))
                                else:
                                    result = str(exp)
                                    logger.info("THERE WAS A PROBLEM RUNNING THE MAIN FUNCTION: %s"%str(result))
                                    
                        else:
                            result = "GROUP TARBALL IS MISSING"
                        
                    else:
            
                        result = "GROUP TARBALL IS MISSING"
            
                    #send the result to minio and close everything down
                    minio_sender = Minio_tool(self.minio_client)
            
            
                    if not isinstance(result, str):
            
                        try:
                            new_payloads = minio_sender.send_file("/output/results.json", meta)
                            
                            if isinstance(result, dict):
                                result['storePayloads'] = new_payloads
                            elif isinstance(result, str):
                                pass
                            else:
                                result = {'storePayloads':new_payloads}
                                
                            #out_dict = {'storePayloads':new_payloads}
                            
                        except Exception as exp:
                            logger.info("FAILED TO SEND RESULT: %s"%str(exp))
                            
                            result = str(exp)
                        
                    #send our results to the next plugin in the queue
                        
                    logger.info("FINISHED RUNNING USER CODE AT %s"%str(datetime.datetime.now()))
                    logger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
    
                    #logger.info(json.dumps(result))
                    #logger.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
                    send(meta, result, self.minio_client, self.debug_mode)
                
                self.backoff = 0
                
            elif code == 404:
                time.sleep(2)
                if self.new:
                    self.new = False
                    self.backoff = 0
                    logger.warning("CONNECTED TO CONTROLER")
                    
            #this controls the backoff if we can't connect
            else:
                if self.backoff == 0:
                    logger.warning("FAILED TO CONNECT TO CONTROLLER, STARTING BACKOFF")
                else:
                    logger.warning("TRIED TO CONNECT AGAIN AND FAILED, retrying in {0}s".format(min(self.backoff + 1, 30)))
                self.backoff = min(self.backoff + 1, 30)
                
                time.sleep(self.backoff)
                
                self.new = False

#check the meta to see if we're working with groups
def check_for_group(meta):
    
    if not isinstance(meta, dict):
        raise Exception('meta should be a dictionary, is actually: %s, a %s'%(str(meta), str(type(meta))))
    
    try:
        result = (meta['job']['workflow']['type'] == 'GROUP')
    except:
        result = False
        
    return result

#download a tarball from a url
def pull_tarball_url(meta):
    
    
    if not isinstance(meta, dict):
        raise Exception('meta should be a dictionary, is actually: %s, a %s'%(str(meta), str(type(meta))))
    
    try:
        url = meta['task']['metadata']['cacheURL']
    except:
        url = None
        
    return url


#This function handles all the tarball stuff
def pull_and_extract_tarball(tar_url, cache_folder_name):
    
    logger.info("DRAWING TARBALL FROM URL")
    
    if not isinstance(cache_folder_name, str):
        raise Exception("The cache folder should be a creatable path, is actually: %s, a %s"%(str(cache_folder_name), str(type(cache_folder_name))))
    
    try:
        #different libraries for interacting with urls in python 2 and python 3
        if sys.version_info.major >= 3:
            file_tmp = urllib.request.urlretrieve(tar_url, filename=None)[0]
        else:
            file_tmp = urllib.urlretrieve(tar_url, filename=None)[0]
        
    except Exception as exp:
        raise Exception("COULD NOT FIND TARBALL AT: %s, due to %s"%(tar_url, str(exp)))
    
    #base_name = os.path.basename(tar_url)
    

    #except Exception as e:
    #    logger.info("COULD NOT PULL TARBALL BECAUSE: %s"%str(e))
    
    tar = tarfile.open(file_tmp)
    
    logger.info("EXTRACTING TARBALL")
    tar.extractall(cache_folder_name)
    

def read_jsonld(filename):
    
    if not isinstance(filename, str):
        
        raise Exception("Filename must be a string, is actually %s, a %s"%(str(filename), str(type(filename))))
    
    f = open(filename, 'r')
    raw = f.read()
    f.close()
    
    return json.loads(raw)


class writer():
    
    def __init__(self, meta):
        
        if not isinstance(meta, dict):
            raise Exception("metadata should be in a dictionary, is actually: %s, a %s"%(str(meta), str(type(meta))))
        
        self.meta = meta
        self.out_folder = '/output'
        self.output_filename = 'results.json'
        self.out_file = os.path.join(self.out_folder, self.output_filename)
        self.initialise_output_file()
        
    def initialise_output_file(self):
    
        if not os.path.isdir(self.out_folder):
            os.mkdir(self.out_folder)
    
        logger.info("Creating output file")
        

        f = open(os.path.join(self.out_folder, self.output_filename), "w")
        f.write('')
        f.close()
            
    #create a jsonld for the whole group job
    def add_group_jsonld(self, group_jsonld):
    
        if group_jsonld != {}:
            #the group jsonld needs to be labeled with a line thats a json formatted like this
            group_line_dict = {}
            group_line_dict['update'] = {}
            group_line_dict['update']['_id'] = self.meta['task']['_id']
            group_line_dict['update']['_type'] = 'groupResults'
            group_line_dict['update']['_index'] = 'group'
            
            
            group_store_dict = {}
            group_store_dict['doc_as_upsert'] = True
            group_store_dict['doc'] = {}
            group_store_dict['doc']['meta'] = {}
            group_store_dict['doc']['meta']['userId'] = self.meta['task']['userId']
            group_store_dict['doc']['meta']['projectId'] = self.meta['task']['projectId']
            group_store_dict['doc']['meta']['jobId'] = self.meta['task']['jobId']
            group_store_dict['doc']['meta']['taskId'] = self.meta['task']['_id']
            group_store_dict['doc']['meta']['storedAt'] = datetime.datetime.utcnow().isoformat()
            group_store_dict['doc']['jsonLD'] = group_jsonld
            
            
            with open(self.out_file, "r+") as f:
                
                lines = f.readlines()
                
                f.seek(0)
               
                f.write(json.dumps(group_line_dict) + '\n')
        
                f.write(json.dumps(group_store_dict) + '\n')
    
                f.writelines(lines)
    
    #create a jsonld for the info added to a given single file. This might be the
    #topic modeling results for this specific single file for example
    def append_task(self, single_file):
        
        if sys.version_info.major >= 3:
            if 'nanowire_plugin.group_tools.single_file' not in str(type(single_file)):
                raise Exception("You can only write a nanowire plugin single_file object to the output using the append task command. You have tried to send an invalid %s object"%str(type(single_file)))
        else:
            #logger.info(str(single_file))
            if 'nanowire_plugin.group_tools.single_file' not in str(single_file):
                raise Exception("You can only write a nanowire plugin single_file object to the output using the append task command. You have tried to send an invalid %s object"%str(type(single_file)))
 
        #we only save a single file result if there have been changes
        if single_file.change_dict != {}:    
        
            #Each single file jsonld needs to be labeled like this
            task_line = {}
            task_line['update'] = {}
            #label the jsonld
            task_line['update']['_id'] = single_file.filename.split("/")[-1] + ':' + self.meta['task']['_id']
            task_line['update']['_type'] = 'taskResults'
            task_line['update']['_index'] = 'group'
            task_line['update']['_parent'] = self.meta['task']['_id']
            
            #This is how we store the individual jsonlds
            task_store_line = {}
            task_store_line['doc'] = single_file.change_dict
            task_store_line['doc_as_upsert'] = True
        
            with open(self.out_file, 'a') as f:
                
                f.write(json.dumps(task_line) + '\n')
                f.write(json.dumps(task_store_line) + '\n')
                

class reader():
    
    def __init__(self):
        self.file_cache = '/cache/jsonlds'
        self.files = os.listdir(self.file_cache)
    #a function to create a generator to pull data
    def file_generator(self):
        
        for file_dat in self.files:
            
            filename = os.path.join(self.file_cache, file_dat)
            yield single_file(filename)

class Minio_tool():
    
    def __init__(self, minio_client):
        
        if "minio.api.Minio" not in str(type(minio_client)):
            
            raise Exception("Minio_tool requires a minio client to initialise, has actually been given: %s, a %s"%(str(minio_client), str(type(minio_client))))
        
        self.minioClient = minio_client
        
        
    def send_file(self, filename, meta):
        
        
        try:
            if 'MINIO_BUCKET' in os.environ.keys():
                bucket_name = os.environ['MINIO_BUCKET']
            else:
                bucket_name = meta['job']['_id']
            job_id = meta['job']['_id']
            task_id = meta['task']['_id']
            
        except Exception as e:
            logger.warning(str(e))
            raise Exception("Key information missing from metadta either job_id or task_id. metadata is: %s"%json.dumps(meta))
        
        logger.info("ENSURING EXISTANCE OF BUCKET: %s"%bucket_name)
        #first check the bucket we want to save to exists
        if not self.minioClient.bucket_exists(bucket_name):
            
            self.minioClient.make_bucket(bucket_name)
            
        save_name = '%s/%s/group.bin'%(job_id, task_id)
        
        #read the outputted jsonld for storage
        logger.info("READING SAVE FILE")
        if not os.path.exists(filename):
            raise Exception("Tried to send non-existant file: %s"%filename)
        
        file_stat = os.stat(filename)
        file_data = open(filename, 'rb')
        
        #send the outputted jsonld to minio
        logger.info("PUTTING OBJECT")
        self.minioClient.put_object(bucket_name, save_name, file_data, file_stat.st_size)
        
        #remove the cache from the pod
        logger.info("CLEANING UP")
        self.clean_up_after_sending()
        
        #add the minio storage point to the metadata
        
        
        try:
            new_store_payloads = meta['task']['metadata']['storePayloads']
            new_store_payloads.append(save_name)
        except:
            new_store_payloads = [save_name]

        
        return new_store_payloads
        
        
    def clean_up_after_sending(self):
        
        if os.path.exists('/cache'):
            shutil.rmtree('/cache')
            
        if os.path.exists('/output'):
            shutil.rmtree('/output')
        

#use this class to make sure all the data from each file stays together
class single_file():
    
    def __init__(self, filename):
        
        file_cache = '/cache' 
        
        if not os.path.exists(os.path.join(file_cache, filename)):
            raise Exception("File to be loaded does not exist: %s"%filename)

        self.filename = filename
        self.jsonld = read_jsonld(os.path.join(file_cache, filename))
        self.change_dict = {}
        
        
def run_group_function(function , read_tool, write_tool, meta):
    
    arguments = inspect.getargspec(function)[0]
    
    #3 argument calls
    if arguments == ['reader', 'writer', 'meta'] or arguments == ['self', 'reader', 'writer', 'meta']:
        
        out = function(read_tool, write_tool, meta)
        
        return out
        
    if arguments == ['reader', 'meta', 'writer'] or arguments == ['self', 'reader', 'meta', 'writer']:
        
        out = function(read_tool, meta, write_tool)
        
        return out
        
    elif arguments == ['meta', 'reader', 'writer'] or arguments == ['self', 'meta', 'reader', 'writer']:
        
        out = function(meta, read_tool, write_tool)
        
        return out
        
    elif arguments == ['meta', 'writer', 'reader'] or arguments == ['self', 'meta', 'writer', 'reader']:
        
        out = function(meta, write_tool, read_tool)
        
        return out
        
    elif arguments == ['writer', 'reader', 'meta'] or arguments == ['self', 'writer', 'reader', 'meta']:
        
        out = function(write_tool, read_tool, meta)
        
        return out
        
    elif arguments == ['writer', 'meta', 'reader'] or arguments == ['self', 'writer', 'meta', 'reader']:
        
        out = function(write_tool, meta, read_tool)
        
        return out
    
    
    #2 arguments calls
    elif arguments == ['reader', 'writer'] or arguments == ['self', 'reader', 'writer']:
        
        out = function(read_tool, write_tool)
        
        return out
        
    elif arguments == ['writer', 'reader'] or arguments == ['self', 'writer', 'reader']:
        
        out = function(write_tool, read_tool)
        
        return out
        

    else:
        
        raise Exception("FUNCTION MUST ACCEPT VALID ARGUMENTS, CURRENT ARGUMENTS ARE %s"%str(arguments))



def group_bind(function, version="1.0.0", debug_mode=0):
    """binds a function to the input message queue"""
    
    #set up the logging
    logger.setLevel(logging.DEBUG)
    
    #pools.set_limit(None)


    #write to screen to ensure logging is working ok
    #print "Initialising nanowire lib, this is a print"
    logger.info("initialising nanowire lib")

    
    #set up the minio client
    minio_client = Minio(
        environ["MINIO_HOST"] + ":" + environ["MINIO_PORT"],
        access_key=environ["MINIO_ACCESS"],
        secret_key=environ["MINIO_SECRET"],
        secure=True if environ["MINIO_SCHEME"] == "https" else False)
        
    #minio_client.set_app_info(name, version)

    monitor_url = environ["MONITOR_URL"]

    worker = GroupWorker(function, minio_client, monitor_url, debug_mode)
    logger.info("Starting consuming")
    worker.run()
                


def validate_group_function(function):
    
    if sys.version_info.major == 3:
        
        arguments = list(inspect.signature(function).parameters)
      
    elif sys.version_info.major == 2:
        
        arguments = inspect.getargspec(function)[0]
        
    allowed = ['self', 'meta', 'reader', 'writer']
    
    arg_dict = set()
    
    for arg in arguments:
        
        if arg not in arg_dict:
            arg_dict.add(arg)
        else:
            raise Exception("ARGUMENTS MAY NOT BE REPEATED")
            
        if arg not in allowed:
            raise Exception("FUNCTION MAY ONLY USE ALLOWED ARGUMENTS, ALLOWED ARGUMENTS ARE: reader, writer, meta YOU HAVE USED THE ARGUMENT %s"%arg)
    
    if 'reader' not in arguments:
        raise Exception("GROUP ANALYSIS FUNCTION MUST TAKE reader AS AN ARGUMENT. THIS IS A CLASS FOR READING DATA")
        
    if 'writer' not in arguments:
        raise Exception("GROUP ANALYSIS FUNCTION MUST TAKE writer AS AN ARGUMENT. THIS IS A CLASS FOR WRITING RESULTS")
    



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 23 11:33:24 2018

@author: stuart
"""

# single file processing tools for nanowire-plugin

import os
from minio import Minio
import time
import traceback
import logging
import json
from os import environ
import inspect

import requests

# from ssl import PROTOCOL_TLSv1_2


import sys
# import hashlib
import datetime

# import urllib.request

from nanowire_plugin import send

# from minio.error import AccessDenied


###############################################################################
### Here the tools that could be exported to single file plugin can be sent ###
###############################################################################
#!!!!!!!!!!!!!!!!!!!!!!!!

# set up the logger globally
logger = logging.getLogger("nanowire-plugin")

logging.getLogger("urllib").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

logging.basicConfig(level=0, stream=sys.stdout)

class Worker(object):
    def __init__(self, function, debug_mode, monitor_url, minio_client):

        self.function = function
        self.minio_client = minio_client
        self.monitor_url = monitor_url
        self.debug_mode = debug_mode
        self.backoff = 0
        self.new = True

        logger.info("ESTABLISHED WORKER")

    def run(self):
        
        logger.info("STARTING HOLD LOOP")
        
        while True:
            
            try:
                
                if 'REQUEST_TIMEOUT' in os.environ.keys():
                    time_out = int(os.environ['REQUEST_TIMEOUT'])
                else:
                    time_out = 5
                    
                message = requests.get(os.environ['CONTROLLER_BASE_URI'] + '/v1/tasks/?pluginId=' +
                                       os.environ['PLUGIN_ID'] + '&pluginInstance=' + os.environ['POD_NAME'], timeout=time_out)
                
                #logger.info("-------------------")
                #logger.info(message.status_code)
                #logger.info(dir(message))
                #logger.info(message.text)
                #logger.info("-------------------")
                
                code = message.status_code
                
            except:
                code = 500

            if code == 200:
                self.backoff = 0
                payload = json.loads(message.text)

                meta = payload['metadata']
                jsonld = payload['jsonld']
                try:
                    url = payload['url']
                except:
                    url = None

                if meta['job']['workflow']['type'] == 'GROUP':
                    logger.warning(
                        "SINGLE FILE PLUGIN TOOL WAS SENT A GROUP JOB")

                else:
                    # try to run our function
                    try:
                        # result = self.function(metadata, jsonld, url)
                        result = run_function(self.function, meta, jsonld, url)

                    except Exception as exp:
                        if self.debug_mode > 0:
                            result = str(traceback.format_exc())
                            logger.info(
                                "THERE WAS A PROBLEM RUNNING THE MAIN FUNCTION: %s" % result)
                        else:
                            result = str(exp)
                            logger.info(
                                "THERE WAS A PROBLEM RUNNING THE MAIN FUNCTION: %s" % result)

                    job_stats = send(
                        meta, result, self.minio_client, self.debug_mode)

                    if self.debug_mode >= 1:
                        logger.info(job_stats)

                    logger.info("FINISHED RUNNING USER CODE AT %s" %
                                str(datetime.datetime.now()))
                    logger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
                    
                    self.new = True

            #404 backoff
            elif code == 404:
                time.sleep(2)
                
                if self.new:
                    
                    logger.info("CONNECTED TO CONTROLER")
                    self.new = False
                    
                self.backoff = 0
                
            #this controls the backoff if we can't connect
            else:
                if self.backoff == 0:
                    logger.info("FAILED TO CONNECT TO CONTROLLER, STARTING BACKOFF")
                    self.new = True
                else:
                    logger.info("TRIED TO CONNECT AGAIN AND FAILED, retrying in {0}s".format(min(self.backoff + 1, 30)))
                    
                self.backoff = min(self.backoff + 1, 30)
                
                time.sleep(self.backoff)


def bind(function, debug_mode=0):
    """binds a function to the input message queue"""

    # set up the logging
    logger.setLevel(logging.DEBUG)

    if debug_mode > 0:
        # write to screen to ensure logging is working ok
        # print "Initialising nanowire lib, this is a print"
        logger.info("Running on %s" % sys.platform)

    # set up the minio client. Do this before the AMPQ stuff

    if 'MINIO_REGION' in environ.keys():
        minio_client = Minio(
            environ["MINIO_HOST"] + ":" + environ["MINIO_PORT"],
            access_key=environ["MINIO_ACCESS"],
            secret_key=environ["MINIO_SECRET"],
            region=environ['MINIO_REGION'],
            secure=True if environ["MINIO_SCHEME"] == "https" else False)
    else:
        minio_client = Minio(
            environ["MINIO_HOST"] + ":" + environ["MINIO_PORT"],
            access_key=environ["MINIO_ACCESS"],
            secret_key=environ["MINIO_SECRET"],
            region=None,
            secure=True if environ["MINIO_SCHEME"] == "https" else False)

    '''
    # use a boto client to try and avoid some serious bugs with minio
    session = boto3.session.Session()
    
    minio_client = session.client('s3', region_name='ams3', endpoint_url=environ['MINIO_HOST']+":"+environ['MINIO_PORT'],
                                  aws_access_key_id=environ['MINIO_ACCESS'],
                                    aws_secret_access_key=environ['MINIO_SECRET'],
                                    secure=True if environ['MINIO_SCHEME']=='https' else False)    
    '''
    # minio_client.set_app_info(name, '1.0.0')

    monitor_url = environ["MONITOR_URL"]

    logger.info("initialised nanowire lib", extra={
        "monitor_url": monitor_url,
        "minio": environ["MINIO_HOST"]
    })

    logger.info("monitor_url: %s" % monitor_url)
    logger.info("minio: %s" % environ["MINIO_HOST"])

    # this is only commented out since I'm trying to find the source of these terrible errors

    worker = Worker(function, debug_mode, monitor_url, minio_client)
    worker.run()
    logger.critical("PAST THE RUN FUNCTION, SOMETHING HAS GONE VERY WRONG")


def validate_single_file_function(function):

    if sys.version_info.major == 3:

        arguments = list(inspect.signature(function).parameters)

    elif sys.version_info.major == 2:

        arguments = inspect.getargspec(function)[0]

    allowed = ['self', 'jsonld', 'metadata', 'url']

    arg_dict = set()

    for arg in arguments:

        if arg not in arg_dict:
            arg_dict.add(arg)
        else:
            raise Exception("ARGUMENTS MAY NOT BE REPEATED")

        if arg not in allowed:
            raise Exception(
                "FUNCTION MAY ONLY USE ALLOWED ARGUMENTS, ALLOWED ARGUMENTS ARE: jsonld, metadata, url, YOU HAVE USED THE ARGUMENT %s" % arg)

    if 'jsonld' not in arguments:

        raise Exception("FUNCTION MUST TAKE jsonld AS AN ARGUMENT")


def run_function(function, metadata, jsonld, url):

    if sys.version_info.major == 3:

        arguments = inspect.signature(function).parameters

    elif sys.version_info.major == 2:

        arguments = inspect.getargspec(function)[0]

    arguments = inspect.getargspec(function)[0]

    # 3 argument variations
    if arguments == ['metadata', 'jsonld', 'url'] or arguments == ['self', 'metadata', 'jsonld', 'url']:

        result = function(metadata, jsonld, url)

        return result

    elif arguments == ['metadata', 'url', 'jsonld'] or arguments == ['self', 'metadata', 'url', 'jsonld']:

        result = function(metadata, url, jsonld)

        return result

    elif arguments == ['jsonld', 'metadata', 'url'] or arguments == ['self', 'jsonld', 'metadata', 'url']:

        result = function(jsonld, metadata, url)

        return result

    elif arguments == ['jsonld', 'url', 'metadata'] or arguments == ['self', 'jsonld', 'url', 'metadata']:

        result = function(jsonld, url, metadata)

        return result

    elif arguments == ['url', 'metadata', 'jsonld'] or arguments == ['self', 'url', 'metadata', 'jsonld']:

        result = function(url, metadata, jsonld)

        return result

    elif arguments == ['url', 'jsonld', 'metadata'] or arguments == ['self', 'url', 'jsonld', 'metadata']:

        result = function(url, jsonld, metadata)

        return result

    # 2 argument variations
    elif arguments == ['url', 'jsonld'] or arguments == ['self', 'url', 'jsonld']:

        result = function(url, jsonld)

        return result

    elif arguments == ['jsonld', 'url'] or arguments == ['self', 'jsonld', 'url']:

        result = function(jsonld, url)

        return result

    elif arguments == ['metadata', 'jsonld'] or arguments == ['self', 'metadata', 'jsonld']:

        result = function(metadata, jsonld)

        return result

    elif arguments == ['jsonld', 'metadata'] or arguments == ['self', 'jsonld', 'metadata']:

        result = function(jsonld, metadata)

        return result

    # 1 argument variations
    elif arguments == ['jsonld'] or arguments == ['self', 'jsonld']:

        result = function(jsonld)

        return result

    else:

        raise Exception(
            "FUNCTION MUST ACCEPT VALID ARGUMENTS, CURRENT ARGUMENTS ARE %s" % str(arguments))



# -*- coding: utf-8 -*-
import requests
import sys
import time
import os
import datetime
import uuid
from os.path import join
from os import environ
import json
import logging
import traceback
"""
Created on Wed Oct 25 11:30:46 2017

@author: stuart
"""

#!/usr/bin/env python3
"""
Provides a `bind` function to plugins so they can simply bind a function to a queue.
"""


#from ssl import PROTOCOL_TLSv1_2

#from minio import Minio

# import the relavant version of urllib depending on the version of python we are
if sys.version_info.major == 3:
    import urllib
elif sys.version_info.major == 2:
    import urllib
    import urllib2
else:
    import urllib


#import hashlib

# set up the logger globally
logger = logging.getLogger("nanowire-plugin")


##################################################################
### These tools are used by both group and single file plugins ###
##################################################################

def validate_payload(payload):
    """ensures payload includes the required metadata and this plugin is in there"""

    if not isinstance(payload, dict):
        raise Exception("payload is a %s, not a dictionary" % type(payload))

    if "nmo" not in payload:
        raise Exception("No nmo in payload")

    if "job" not in payload["nmo"]:
        raise Exception("No job in nmo \nnmo is %s" % payload["nmo"])

    if "task" not in payload["nmo"]:
        raise Exception("No task in nmo \nnmo is %s" % payload["nmo"])

    try:
        isGroup = payload['nmo']['source']['misc']['isGroup']
    except:
        isGroup = False

    if "jsonld" not in payload and not isGroup:
        raise Exception("No jsonld in payload \nPayload is:- %s" % payload)


# Send the data back to the controler
def send(metadata, output, minio_client, debug_mode):
    """unwraps a message and calls the user function"""

    # log some info about what the send function has been given
    logger.info("LETTING MONITOR KNOW PROCESSING HAS BEEN DONE")

    if isinstance(output, str):
        err = output
        output = {}
    elif isinstance(output, dict):
        err = None
    elif output == None:
        output = {}
        err = "NO OUTPUT WAS RETURNED"

    # send the info from this plugin to the next one in the pipeline
    send_result(output, metadata, err)


def send_result(output, metadata, error):

    logger.info("RUNNING SEND TO NEXT PLUGIN")

    # check if we're dealing with a group job. If it is a group job
    # any additional metadata created will be stored in what is erroneously
    # labeled as jsonld to make life easier
    if metadata['job']['workflow']['type'] == 'GROUP':

        # Group job went without a hitch
        if error == None:

            logger.debug("NO ERRORS DETECTED RETURNING FINE JSONLD")

            send_json = {
                "pluginInstance": os.environ["POD_NAME"],
                "status": "success",
                "additionalMetadata": output
            }
        # this is the instance of we had an error when calculating the group
        # job
        else:

            logger.debug("ERROR WAS DETECTED")
            logger.debug(error)

            send_json = {
                "pluginInstance": os.environ["POD_NAME"],
                "status": "failure",
                'error': error,
                "additionalMetadata": output
            }

            # logger.debug("~~~~~~~~~~~~~~~~~~~~~~~~~")
            # logger.debug(json.dumps(send_json))

    # if we don't have a group job we must have a single file job.
    else:

        if error == None:

            send_json = {'pluginInstance': os.environ['POD_NAME'],
                         'status': 'success',
                         'jsonld': output}

        else:
            # logger.warning(error)
            # logger.warning("++++++++++++++++++++++")
            send_json = {'pluginInstance': os.environ['POD_NAME'],
                         'status': 'failure',
                         'error': error,
                         'jsonld': output}

    # This bit sends the result to the controller. It is the most likely
    # source of trouble
    trying = True
    tries = 0
    backoff = 1
    while trying:
        try:
            r = requests.put(os.environ['CONTROLLER_BASE_URI'] + '/v1/tasks/' + metadata['task']['_id'],
                             data=json.dumps(send_json), headers={'content-type': 'application/json'})
            trying = False
        except:
            tries += 1
            time.sleep(backoff)
            backoff = min(30, backoff * 2)
            logger.warning("%s ATTEMPTS MADE TO PUT TASK RESULT" % str(tries))

    if r.status_code != 200:
        logger.warning(str(dir(r)))
        logger.warning(str(r.reason))
        logger.warning(str(r.text))
        logger.warning("-----------")
        logger.warning(str(output))
        logger.warning(str(type(output)))
        logger.warning("---------")
        logger.warning("FAILED PUT WITH STATUS: %s" % str(r.status_code))


def grab_file(url):

    if not os.path.exists('/tmp'):
        os.mkdir('/tmp')
    try:
        filename = '/tmp/document'
        urllib.request.urlretrieve(url, filename)
        return filename
    except:

        return "COULD NOT GRAB DOCUMENT, URL MAY HAVE EXPIRED"


# EXPEREMENTAL SECTION
def create_task(nmo, new_jsonld):

    job_id = nmo["job"]["job_id"],

    monitor_url = environ["MONITOR_URL"]

    task_dict = {}
    task_dict['_id'] = str(uuid.uuid4())
    task_dict["name"] = "Extracted data file"
    task_dict["type"] = "TestAction"
    task_dict["jobId"] = job_id

    upload_dict = {}
    upload_dict["type"] = "Example"
    upload_dict["meta"] = new_jsonld
    upload_dict["credentials"] = {"accessToken": environ["AMQP_USER"],
                                  "accessTokenSecret": environ["AMQP_PASS"],
                                  "consumerKey": environ["AMQP_USER"],
                                  "consumerSecret": environ["AMQP_PASS"]}
    payload = json.dumps({
        "task": task_dict,
        "upload": upload_dict})

    # if we're working with python3
    if sys.version_info.major == 3:

        # if debug_status >= 2:
        #    logger.info("Running in python 3")

        request_url = monitor_url + "/v1/tasks"
        # urllib.parse.urljoin(monitor_url,"/v4/tasks/%s/positions"%task_id)

        req = urllib.request.Request(request_url,
                                     payload.encode(),
                                     headers={
                                         "Content-Type": "application/json"
                                     })

        result = urllib.request.urlopen(req)

        #logger.info("Task creation result was")
        # logger.info(str(result))

    # if we're working with python2
    elif sys.version_info.major == 2:

        # if debug_status >= 2:
        #    logger.info("Running in python 2")

        # there's no urljoin command in python2
        request_url = monitor_url + "/v1/tasks"

        req = urllib2.Request(request_url,
                              payload.encode(),
                              headers={
                                  "Content-Type": "application/json"
                              })

        result = urllib2.urlopen(req)

        logger.info("Task creation result was")
        logger.info(str(result))

    # if we're not in python2 or python3
    else:

        logger.warning("Running in an unknown version of python:- %s" %
                       str(sys.version_info))
